from __future__ import annotations


# standard library
import asyncio
from dataclasses import dataclass
from datetime import datetime, timedelta
from logging import getLogger
from typing import AsyncIterable, Awaitable, Iterable, Optional, Sequence, Union


# third-party packages
from aiohttp import ClientSession, TCPConnector
from feedparser import FeedParserDict, parse
from typing_extensions import Final
from .article import Article


# constants
ARXIV_API: Final[str] = "http://export.arxiv.org/api/query"
DATE_FORMAT: Final[str] = "%Y%m%d%H%M%S"
N_CONCURRENT: Final[int] = 1
N_MAX_ARTICLES: Final[int] = 1000
N_PER_REQUEST: Final[int] = 100
SECOND: Final[timedelta] = timedelta(seconds=1)


# module-level logger
logger = getLogger(__name__)


# data classes
@dataclass
class Search:
    """Search class for searching for articles in arXiv."""

    date_start: Union[datetime, str]  #: Start date for a search (inclusive).
    date_end: Union[datetime, str]  #: End date for a search (exclusive).
    keywords: Optional[Sequence[str]] = None  #: Keywords for a search.
    categories: Optional[Sequence[str]] = None  #: arXiv categories.
    n_max_articles: int = N_MAX_ARTICLES  #: Maximum number of articles to get.
    n_per_request: int = N_PER_REQUEST  #: Number of articles to get per request.
    n_concurrent: int = N_CONCURRENT  #: Number of simultaneous requests.

    def __post_init__(self) -> None:
        if not isinstance(self.date_start, datetime):
            self.date_start = datetime.fromisoformat(self.date_start)

        if not isinstance(self.date_end, datetime):
            self.date_end = datetime.fromisoformat(self.date_end)

    @property
    def search_query(self) -> str:
        """Convert to search query for the arXiv API."""
        date_start = self.date_start.strftime(DATE_FORMAT)
        date_end = (self.date_end - SECOND).strftime(DATE_FORMAT)

        query = f"submittedDate:[{date_start} TO {date_end}]"

        if self.categories:
            sub = " OR ".join(f"cat:{cat}" for cat in self.categories)
            query += f" AND ({sub})"

        if self.keywords:
            sub = " OR ".join(f"abs:{kwd}" for kwd in self.keywords)
            query += f" AND ({sub})"

        return query

    def __aiter__(self) -> AsyncIterable[Article]:
        """Search for articles and yield them as article instances."""

        async def search():
            async for entry in self._search():
                yield Article(
                    title=entry.title,
                    authors=[author.name for author in entry.authors],
                    summary=entry.summary,
                    arxiv_url=entry.link,
                )

        return search()

    async def _search(self) -> AsyncIterable[FeedParserDict]:
        """Search for articles and yield them as Atom entries."""
        connector = TCPConnector(limit=self.n_concurrent)

        async with ClientSession(connector=connector) as client:
            requests = list(self._gen_requests(client))

            for request in asyncio.as_completed(requests):
                feed = parse(await request)

                for entry in feed.entries:
                    yield entry

    def _gen_requests(self, client: TCPConnector) -> Iterable[Awaitable]:
        """Generate coroutines to request the arXiv results."""

        async def request(url: str, **params) -> Awaitable[str]:
            async with client.get(url, params=params) as resp:
                logger.debug(resp.url)
                return await resp.text()

        for start in range(0, self.n_max_articles, self.n_per_request):
            max_results = min(
                self.n_per_request,
                self.n_max_articles - start,
            )

            yield request(
                ARXIV_API,
                search_query=self.search_query,
                start=start,
                max_results=max_results,
            )


# utility functions
def search(
    date_start: Union[datetime, str],
    date_end: Union[datetime, str],
    keywords: Optional[Sequence[str]] = None,
    categories: Optional[Sequence[str]] = None,
    n_max_articles: int = N_MAX_ARTICLES,
    n_per_request: int = N_PER_REQUEST,
    n_concurrent: int = N_CONCURRENT,
) -> Sequence[Article]:
    """Search for articles in arXiv with given conditions.

    Args:
        date_start: Start date for a search (inclusive).
        date_end: End date for a search (exclusive).
        keywords: Keywords for a search.
        categories: arXiv categories.
        n_max_articles: Maximum number of articles to get.
        n_per_request: Number of articles to get per request.
        n_concurrent: Number of simultaneous requests.

    Returns:
        List of articles found in arXiv.

    """
    search = Search(
        date_start,
        date_end,
        keywords,
        categories,
        n_max_articles,
        n_per_request,
        n_concurrent,
    )

    async def coro() -> Awaitable:
        return [article async for article in search]

    return asyncio.run(coro())



from __future__ import annotations


# standard library
import re
from dataclasses import dataclass, replace
from typing import Pattern, Optional, Sequence, Union


# third-party packages
from typing_extensions import Final


# constants
SEPARATOR: Final[str] = "++++++++++"


# data classes
@dataclass
class Article:
    """Article class for storing article information."""

    title: str  #: Title of an article.
    authors: Sequence[str]  #: Author(s) of an article.
    summary: str  #: Summary of an article.
    arxiv_url: str  #: arXiv URL of an article.

    def __post_init__(self) -> None:
        self.title = detex(self.title)
        self.summary = detex(self.summary)

    def replace(self, text: str, translated: str) -> Article:
        """Method necessary to become translatable."""
        title, summary = translated.split(SEPARATOR)
        return replace(self, title=title, summary=summary)

    def __str__(self) -> str:
        """Method necessary to become translatable."""
        return f"{self.title}\n{SEPARATOR}\n{self.summary}"


@dataclass
class Rule:
    """Rule class for replacing one string with another."""

    pattern: Union[Pattern, str]  #: Text or pattern for search.
    replacement: str  #: Text or pattern for replacement.

    def __post_init__(self):
        if isinstance(self.pattern, str):
            self.pattern = re.compile(self.pattern)

    def apply(self, text: str) -> str:
        """Apply the rule to text."""
        return self.pattern.sub(self.replacement, text)


# utility functions
default_rules = [
    Rule(r"\\text(bb|bf|it|gt|mc|md|rm|sc|sf|sl|tt|up){(.+?)}", r"\2"),
    Rule(r"{\\text(bb|bf|it|gt|mc|md|rm|sc|sf|sl|tt|up) (.+?)}", r"\2"),
    Rule(r"{\\(bf|em|it|rm|sc|sf|sl|tt) (.+?)}", r"\2"),
    Rule(r"\\emph{(.+?)}", r"\1"),
    Rule(r"(\n+\s*|\n*\s+)", " "),
]


def detex(text: str, rules: Optional[Sequence[Rule]] = None) -> str:
    """Remove TeX's control commands from text."""
    if rules is None:
        rules = default_rules

    for rule in rules:
        text = rule.apply(text)

    return text



# standard library
import asyncio
from datetime import date, datetime, time, timedelta
from logging import basicConfig, DEBUG, getLogger
from typing import Union


# third-party packages
from fire import Fire
from typing_extensions import Final
from .apps import Slack
from .search import Search
from .translate import DeepL


# constants
CATEGORIES: Final[str] = (
    "astro-ph.CO,"
    "astro-ph.EP,"
    "astro-ph.GA,"
    "astro-ph.HE,"
    "astro-ph.IM,"
    "astro-ph.SR"
)
KEYWORDS: Final[str] = ""
LANG_FROM: Final[str] = "en"
LANG_TO: Final[str] = "auto"
N_CONCURRENT: Final[int] = 10
TIMEOUT: Final[int] = 60


# module-level logger
logger = getLogger(__name__)


# helper functions
def n_days_ago(n: int) -> str:
    """Return string of date n days ago in YYYY-mm-dd."""
    today = datetime.combine(date.today(), time())
    return (today - timedelta(days=n)).strftime("%Y-%m-%d")


def parse_multiple(obj: Union[list, tuple, str], sep: str = ",") -> list:
    """Parse comma-separated multiple values correctly."""

    def remove_empty(strings: list) -> list:
        return [s for s in strings if s != ""]

    if isinstance(obj, str):
        return remove_empty(obj.split(sep))

    if isinstance(obj, (list, tuple)):
        return remove_empty(sep.join(obj).split(sep))

    raise ValueError(f"Invalid value: {obj:!r}")


def configure_logging(debug: bool = False) -> None:
    """Configure logging format and level for CLI."""
    basicConfig(
        datefmt="%Y-%m%d %H:%M:%S",
        format="[{asctime} {name} {levelname}]: {message}",
        style="{",
    )

    if debug:
        getLogger("astro_ph").setLevel(DEBUG)


# subcommand functions
def slack(
    date_start: str = n_days_ago(2),
    date_end: str = n_days_ago(1),
    keywords: str = KEYWORDS,
    categories: str = CATEGORIES,
    lang_from: str = LANG_FROM,
    lang_to: str = LANG_TO,
    timeout: int = TIMEOUT,
    n_concurrent: int = N_CONCURRENT,
    webhook_url: str = "",
    debug: bool = False,
) -> None:
    """Translate and post articles to Slack.

    Args:
        date_start: Start date for a search in YYYY-mm-dd (inclusive).
        date_end: End date for a search in YYYY-mm-dd (exclusive).
        keywords: Comma-separated keywords for a search.
        categories: Comma-separated arXiv categories.
        lang_from: Language of original text in articles.
        lang_to: Language for translated text in articles.
        timeout: Timeout for each post execution (in seconds).
        n_concurrent: Number of simultaneous execution.
        webhook_url: URL of Slack incoming webhook.
        debug: If True, debug-level log messages are shown.

    Returns:
        This command does not return anything.

    """
    configure_logging(debug)

    if not webhook_url:
        raise ValueError("Webhook URL must be specified.")

    keywords = parse_multiple(keywords)
    categories = parse_multiple(categories)

    articles = Search(date_start, date_end, keywords, categories)
    translator = DeepL(lang_from, lang_to)
    app = Slack(translator, n_concurrent, timeout, webhook_url)

    logger.debug(articles)
    logger.debug(translator)
    logger.debug(app)

    asyncio.run(app.post(articles))


# command line interface
def cli() -> None:
    """Entry point of command line interface."""
    Fire(dict(slack=slack))



# standard library
import asyncio
from dataclasses import dataclass
from logging import getLogger
from typing import (
    AsyncIterable,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    Optional,
    TypeVar,
)


# third-party packages
from aiohttp import ClientSession
from typing_extensions import Final, Protocol
from .translate import Translator
from .search import Article


# constants
TIMEOUT: Final[int] = 60
N_CONCURRENT: Final[int] = 10


# module-level logger
logger = getLogger(__name__)


# type hints
S = TypeVar("S")
T = TypeVar("T")
Payload = Dict[str, str]


class App(Protocol):
    """Protocol that defines application objects."""

    translator: Translator  #: Translator object.
    n_concurrent: int  #: Number of simultaneous post execution.
    timeout: int  #: Timeout for each post execution.

    async def post(self, articles: AsyncIterable[Article]) -> Awaitable:
        """Translate and post articles to somewhere."""
        ...


# data classes
@dataclass
class Slack:
    """Class for posting articles to Slack by incoming webhook."""

    translator: Translator  #: Translator object.
    n_concurrent: int = N_CONCURRENT  #: Number of simultaneous post execution.
    timeout: int = TIMEOUT  #: Timeout for each post execution (in seconds).
    webhook_url: str = ""  #: URL of Slack incoming webhook.

    async def post(self, articles: AsyncIterable[Article]) -> Awaitable[None]:
        """Translate and post articles to Slack."""
        await amap(self._post, articles, self.n_concurrent, self.timeout)

    async def _post(self, article: Article) -> Awaitable[None]:
        """Translate and post an article to Slack."""
        translated = await self.translator.translate(article)
        payload = self._to_payload(article, translated)

        async with ClientSession() as client:
            async with client.post(self.webhook_url, json=payload) as resp:
                logger.debug(f"{str(resp.url):.300}...")

        await asyncio.sleep(1.0)

    def _to_payload(self, original: Article, translated: Article) -> Payload:
        """Convert article to payload for Slack post."""
        divider = self.divider()
        header = self.header(self.plain_text(translated.title))
        title = self.section(self.mrkdwn(f"*Title:* {original.title}"))
        authors = self.section(self.mrkdwn(f"*Authors:* {', '.join(original.authors)}"))
        summary = self.section(self.mrkdwn(f"*Summary:* {translated.summary}"))
        buttons = self.actions(
            elements=[
                self.button(
                    self.plain_text("View arXiv"),
                    url=original.arxiv_url,
                ),
                self.button(
                    self.plain_text("View PDF"),
                    url=original.arxiv_url.replace("abs", "pdf"),
                ),
            ]
        )

        blocks = [divider, header, title, authors, summary, buttons, divider]
        return dict(text=translated.title, blocks=blocks)

    def __getattr__(self, type: str) -> Callable[[str], Payload]:
        """Generate a function to create elements of Slack payload."""

        def element(text: Optional[str] = None, **params) -> Payload:
            if text is None:
                return dict(type=type, **params)

            return dict(text=text, type=type, **params)

        return element


# helper functions
async def amap(
    afunc: Callable[[S], Awaitable[T]],
    aiterable: AsyncIterable[S],
    n_concurrent: int = N_CONCURRENT,
    timeout: int = TIMEOUT,
) -> Awaitable[Iterable[T]]:
    """Async map function.

    Args:
        afunc: Coroutine function.
        aiterable: Async iterable that yields args of afunc.
        n_concurrent: Number of concurrent execution.
        timeout: Timeout for each afunc execution (in seconds).

    Returns:
        Generator that yields results of afunc.

    """
    queue_in = asyncio.Queue()
    queue_out = asyncio.Queue()

    # step 1: create consumer coroutines
    async def consume() -> Awaitable[None]:
        async def _consume() -> Awaitable[None]:
            arg = await queue_in.get()
            coro = asyncio.wait_for(afunc(arg), timeout)

            try:
                result = await coro
            except asyncio.TimeoutError as err:
                result = err
            finally:
                await queue_out.put(result)
                queue_in.task_done()

        try:
            while True:
                await _consume()
        except asyncio.CancelledError:
            pass

    consumers = []

    for _ in range(n_concurrent):
        consumer = asyncio.create_task(consume())
        consumers.append(consumer)

    # step 2: create register coroutine
    async def register() -> Awaitable[None]:
        async for arg in aiterable:
            await queue_in.put(arg)

        await queue_in.join()

        for consumer in consumers:
            consumer.cancel()

    # step 3: run coroutines concurrently
    await asyncio.gather(register(), *consumers)

    # step 4: return results as a generator
    def results():
        while not queue_out.empty():
            yield queue_out.get_nowait()

    return results()



from __future__ import annotations


# standard library
import asyncio
from dataclasses import dataclass
from enum import auto, Enum
from logging import getLogger
from typing import Awaitable, Union
from urllib.parse import quote


# third-party packages
from pyppeteer import launch
from typing_extensions import Final, Protocol


# constants
DEEPL_URL: Final[str] = "https://www.deepl.com/translator"
JS_FUNC: Final[str] = "element => element.textContent"
SELECTOR: Final[str] = ".lmt__translations_as_text__text_btn"
TIMEOUT: Final[int] = 60


# module-level logger
logger = getLogger(__name__)


# enums
class Language(Enum):
    """Available languages for translation."""

    AUTO = auto()  #: Auto language detection
    DE = auto()  #: German
    EN = auto()  #: English
    FR = auto()  #: French
    IT = auto()  #: Italian
    JA = auto()  #: Japanese
    ES = auto()  #: Spanish
    NL = auto()  #: Dutch
    PL = auto()  #: Polish
    PT = auto()  #: Portuguese
    RU = auto()  #: Russian
    ZH = auto()  #: Chinese


# type hints
class Translatable(Protocol):
    """Protocol that defines translatable objects."""

    def __str__(self) -> str:
        """Return text to be translated."""
        ...

    def replace(text: str, translated: str) -> Translatable:
        """Replace text with translated one."""
        ...


class Translator(Protocol):
    """Protocol that defines translator objects."""

    lang_from: Union[Language, str]  #: Language of original text.
    lang_to: Union[Language, str]  #: Language for translated text.
    timeout: int  #: Timeout for translation (in seconds).

    async def translate(self, obj: Translatable) -> Awaitable[Translatable]:
        """Translate object written in one language to another."""
        ...


# data classes
@dataclass
class DeepL:
    """DeepL class for translating text."""

    lang_from: Union[Language, str] = Language.AUTO  #: Language of original text.
    lang_to: Union[Language, str] = Language.AUTO  #: Language for translated text.
    timeout: int = TIMEOUT  #: Timeout for translation (in seconds).

    def __post_init__(self) -> None:
        if isinstance(self.lang_from, str):
            self.lang_from = Language[self.lang_from.upper()]

        if isinstance(self.lang_to, str):
            self.lang_to = Language[self.lang_to.upper()]

    @property
    def base_url(self) -> str:
        """Base URL of translation."""
        return f"{DEEPL_URL}#/{self.lang_from.name}/{self.lang_to.name}"

    async def translate(self, obj: Translatable) -> Awaitable[Translatable]:
        """Translate object written in one language to another."""
        if self.lang_from == self.lang_to:
            return obj

        text = str(obj)
        translated = await self._translate_text(text)
        return obj.replace(text, translated)

    async def _translate_text(self, text: str) -> Awaitable[str]:
        """Translate text written in one language to another."""
        url = f"{self.base_url}/{quote(text)}"
        logger.debug(f"{url:.300}...")

        browser = await launch()
        page = await browser.newPage()
        page.setDefaultNavigationTimeout(self.timeout * 1000)
        completion = self._translation_completion(page)

        try:
            await page.goto(url)
            return await asyncio.wait_for(completion, self.timeout)
        except asyncio.TimeoutError as err:
            raise type(err)("Translation was timed out.")
        finally:
            await browser.close()

    async def _translation_completion(self, page) -> Awaitable[str]:
        """Wait for completion of translation and return result."""
        translated = ""

        while not translated:
            await asyncio.sleep(1.0)
            element = await page.querySelector(SELECTOR)
            translated = await page.evaluate(JS_FUNC, element)

        return translated


# utility functions
def translate(
    obj: Translatable,
    lang_to: Union[Language, str] = Language.AUTO,
    lang_from: Union[Language, str] = Language.AUTO,
    timeout: int = TIMEOUT,
) -> Translatable:
    """Translate object written in one language to another.

    Args:
        obj: Object whose text is to be translated.
        lang_to: Language for translated text.
        lang_from: Language of original text.
        timeout: Timeout for translation (in seconds).

    Returns:
        Translated object.

    """
    deepl = DeepL(lang_from, lang_to, timeout)
    return asyncio.run(deepl.translate(obj))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['astro_ph']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7,<4.0',
 'feedparser>=6.0,<7.0',
 'fire>=0.4,<0.5',
 'pyppeteer>=0.2,<0.3',
 'typing-extensions>=3.7,<4.0']

entry_points = \
{'console_scripts': ['astro-ph = astro_ph.cli:cli']}

setup_kwargs = {
    'name': 'astro-ph',
    'version': '0.2.3',
    'description': 'Translate and post arXiv articles to various apps',
    'long_description': '# astro-ph\n\n[![PyPI](https://img.shields.io/pypi/v/astro-ph.svg?label=PyPI&style=flat-square)](https://pypi.org/project/astro-ph/)\n[![Python](https://img.shields.io/pypi/pyversions/astro-ph.svg?label=Python&color=yellow&style=flat-square)](https://pypi.org/project/astro-ph/)\n[![Test](https://img.shields.io/github/workflow/status/astropenguin/astro-ph/Test?logo=github&label=Test&style=flat-square)](https://github.com/astropenguin/astro-ph/actions)\n[![License](https://img.shields.io/badge/license-MIT-blue.svg?label=License&style=flat-square)](LICENSE)\n\nTranslate and post arXiv articles to various apps\n\n## Installation\n\nUse pip or other package manager to install the Python package.\n\n```shell\n$ pip install astro-ph\n```\n\n## Usage\n\nAfter installation, command line interface, `astro-ph`, is available, with which you can translate and post arXiv articles to various apps.\nNote that only `slack` app is currently available.\nIn this case, you need to [create a custom Slack app to get an URL of incoming webhook](https://slack.com/help/articles/115005265063-Incoming-webhooks-for-Slack).\n\n```shell\n$ astro-ph slack --keywords galaxy,galaxies \\\n                 --categories astro-ph.GA,astro-ph.IM \\\n                 --lang_to ja \\\n                 --webhook_url https://hooks.slack.com/services/***/***\n```\n\nThe posted article looks like this.\n\n![astro-ph-slack.png](https://raw.githubusercontent.com/astropenguin/astro-ph/master/docs/_static/astro-ph-slack.png)\n\nFor detailed information, see the built-in help by the following command.\n\n```shell\n$ astro-ph slack --help\n```\n\n## Example\n\nIt would be nice to regularly run the command by GitHub Actions.\nHere is a live example in which daily (2 days ago) arXiv articles in [astro-ph.GA](https://arxiv.org/list/astro-ph.GA/new) and [astro-ph.IM](https://arxiv.org/list/astro-ph.IM/new) are posted to different channels of a Slack workspace.\n\n- [a-lab-nagoya/astro-ph-slack: Translate and post arXiv articles to Slack](https://github.com/a-lab-nagoya/astro-ph-slack)\n\n## References\n\n- [fkubota/Carrier-Owl: arxiv--> DeepL --> Slack](https://github.com/fkubota/Carrier-Owl)\n    - The astro-ph package is highly inspired by their work\n- [a-lab-nagoya/astro-ph-slack: Translate and post arXiv articles to Slack](https://github.com/a-lab-nagoya/astro-ph-slack)\n    - A live example using the astro-ph package\n- [pyppeteer/pyppeteer: Headless chrome/chromium automation library (unofficial port of puppeteer)](https://github.com/pyppeteer/pyppeteer)\n    - Used for async Chromium operation\n- [aio-libs/aiohttp: Asynchronous HTTP client/server framework for asyncio and Python](https://github.com/aio-libs/aiohttp)\n    - Used for async article posts to Slack\n- [google/python-fire: Python Fire is a library for automatically generating command line interfaces (CLIs) from absolutely any Python object.](https://github.com/google/python-fire)\n    - Used for creating command line interface\n',
    'author': 'Akio Taniguchi',
    'author_email': 'taniguchi@a.phys.nagoya-u.ac.jp',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/astropenguin/astro-ph/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
